# Class 02 - Functions, Modules, and Packages

Python (like most programming languages) gives you the ability to reuse code by organizing your code into functions, modules, and packages.

## What is a function?

A function is a block of code with a name. Functions usually (but don't have to) take **arguments**.

> ***What's an argument?***  
> An argument is a placeholder variable for a value that is not yet known until later when the function is executed ("called").

Functions are **defined** using the keyword `def` . Here's an example:

```python
def multiply(x, y):
    return x * y
```

*What does this code do?* It creates a new function object called `multiply`. This isn't too different from declaring a plain old variable (e.g. `value = 1`). The magic happens when we invoke ("call") the function with values provided for `x` and `y`, assigning the result to a new variable `answer`:

```python
answer = multiply(18, 19)
print(answer)
# 342
```

So in this case, `x` and `y` only exist within the context of `multiply`. This is called **scope**. Every time we invoke `multiply()`, we recreate `x` and `y` with different values:

```python
print(multiply(4, 3))
# 12
print(multiply(8, 11))
# 88
print(multiply(1156, 913))
# 1055428
```

### Type Hints

Quick tangent that is worth mentioning: type hints. Python allows you to annotate function arguments and return value so you can indicate how you intend the function to be used. As an example, here is our `multiply()` function with type hints:

```python
def multiply(x: int, y: int) -> int:
    return x * y
```

To annotate a parameter, add a colon and the expected type (`x: int`), and to annotate the return value, add an "arrow" and the output type before the final colon (`-> int:`).

If type hints aren't necessary, then why go through the trouble? Type hints are considered "good practice" because they force you to consider how you intend for your function to be used. This is very important if other people will be looking at your code (or as a reminder to yourself in a few months' time).

### Docstrings

Like type hints, a useful but technically optional feature of functions is the **docstring**. A docstring is a simple description of what your function does, explaining any parameters, return values, and side effects (we'll get to side effects later on). Docstrings are meant for humans (not unlike comments), so be as brief or descriptive as you see fit. As an example, let's add a docstring to our `multiply()` function:

```python
def multiply(x: int, y: int) -> int:
    """
    Multiply two numbers together (x and y) and return the product.
    """
```

A more formal docstring you might encounter in the wild may look something like this:

```python
def multiply(x: int, y: int) -> int:
    """
    Parameters
    ----------
    x: int
        The value in the left side of the equation.

    y: int
        The value in the right side of the equation.

    Returns
    -------
    The product of the parameters `x` and `y`.
    """
    return x * y
```

Although it looks a little verbose for such a toy example, it's actually an example of clean code when your docstring is longer than the function body itself!

### Scope

A convenient quirk of Python's objects system is that functions can access variables in their parent scope.

```python
GLOBAL_NUMBERS = (2, 4, 6, 8)

def display_numbers():
    local_numbers = (1, 3, 5, 7, 9)
    print(f"{local_numbers=}")
    print(f"{GLOBAL_NUMBERS=}")

def inspect_numbers(numbers_to_inspect):
    print(f"{numbers_to_inspect=}")
    print(f"{GLOBAL_NUMBERS=}")

display_numbers()
# local_numbers=(1, 3, 5, 7, 9)
# GLOBAL_NUMBERS=(2, 4, 6, 8)

inspect_numbers((9, 8, 7, 6, 5))
# numbers_to_inspect=(9, 8, 7, 6, 5)
# GLOBAL_NUMBERS=(2, 4, 6, 8)
```

In this example, both functions `display_numbers()` and `inspect_numbers()` are able to access the value of `GLOBAL_NUMBERS`.

> It's a convention to give variables upper-case names (e.g. `GLOBAL_NUMBERS`) to signal that a variable is global and intended to be read-only.

### Side Effects

One good reason for docstrings and type hints is to keep track of any **side effects**.

**What is a side effect?**

A side effect is when your function changes the state of anything *beyond its local scope*.

> A function with no side effects (e.g. our `multiply()` example) is called a **pure function**.

Python functions have the ability to access objects in their parent scope. For **immutable** objects (strings, ints, floats, tuples, etc.) this is convenient, but for mutable objects (lists, dicts), it's extremely important to keep track of where your object is allowed to be accessed.

> Immutable means the object cannot be modified. Mutable means the object's state may be changed.

Let's look at an example:

```python
secret_numbers = [1, 2, 3, 5, 8, 13]

def modify_secrets(new_secret: int) -> None:
    secret_numbers.append(new_secret)

modify_secrets(99)
print(secret_numbers)
# [1, 2, 3, 5, 8, 13, 99]
```

In this function, 

## What is a module?

## What is a package?